/*
  全方位を2のべき乗分割する2進角度と
  テーブル参照による高速三角関数
*/
#ifndef BIN_DEG_H
#define BIN_DEG_H


#define PI 3.1415926536

// デフォルトで全方位を何分割するか
// 変更する場合は、以下の3マクロをすべて定義する(要bin_deg.cの再コンパイル)
// BIN_DEG_RESOLUTIONは2のべき乗にすること
// ATAN_RESOLUTIONはあるていど独立して調整可能だが、
// BIN_DEG_RESOLUTIONに応じて大きくしないとbin_deg_atan2の精度が落ちる
#ifndef BIN_DEG_RESOLUTION
//#define BIN_DEG_RESOLUTION 256
//#define BIN_DEG_RESOLUTION_BITS 8
//#define BIN_DEG_ATAN_RESOLUTION 1024
#define BIN_DEG_RESOLUTION 1024
#define BIN_DEG_RESOLUTION_BITS 10
#define BIN_DEG_ATAN_RESOLUTION 1024
#endif

// 高精度角度を使用する場合、下位に何ビット使用するか
// 変更する場合は、以下の2マクロをすべて定義する(要bin_deg.cの再コンパイル)
// たとえばBIN_DEG_RESOLUTION=256, BIN_DEG_SUB_RESOLUTION=1024なら
// bin_deg_tは全方位256分割、bin_hdeg_tは全方位256x1024分割となり、
// 後者は8+10=18ビット長表現になる(10ビット右シフトでbin_deg_tに変換可能)。
#ifndef BIN_DEG_SUB_RESOLUTION
#define BIN_DEG_SUB_RESOLUTION 4096
#define BIN_DEG_SUB_RESOLUTION_BITS 12
#endif

#define BIN_HDEG_RESOLUTION_BITS \
    (BIN_DEG_RESOLUTION_BITS+BIN_DEG_SUB_RESOLUTION_BITS)
#define BIN_DEG_MASK ((1 << BIN_DEG_RESOLUTION_BITS)-1)
#define BIN_HDEG_MASK ((1 << BIN_HDEG_RESOLUTION_BITS)-1)

typedef int bin_deg_t;
typedef int bin_hdeg_t;

// 2進角度を正規化(0以上BIN_DEG_RESOLUTION未満)する
#define BinDegNormalize(d) ((d) & BIN_DEG_MASK)
#define BinHDegNormalize(d) ((d) & BIN_HDEG_MASK)

// 90度
#define BIN_DEG_RANGLE (BIN_DEG_RESOLUTION>>2)
// 180度
#define BIN_DEG_SANGLE (BIN_DEG_RESOLUTION>>1)


extern double _bin_deg_sin_table[BIN_DEG_RESOLUTION];
extern double _bin_deg_tan_table[BIN_DEG_RESOLUTION];
extern int _bin_deg_atan_table[BIN_DEG_ATAN_RESOLUTION+1];

// 2進角度とラジアンの相互変換
#define BinDeg2Rad(d) ((d)*2*PI/BIN_DEG_RESOLUTION)
// 端数部分を四捨五入するため0.5ずらす
#define Rad2BinDeg(r)                                                \
    (BinDegAdd((int)((r)*BIN_DEG_RESOLUTION/(2*PI)+((r<0)?-0.5:.5)), \
               BIN_DEG_RESOLUTION))

#define BinHDeg2DegN(d) ((d) >> BIN_DEG_SUB_RESOLUTION_BITS)
#define BinHDeg2Deg(d)                                          \
    ((d < 0) ?                                                  \
     (-(((d) >> BIN_DEG_SUB_RESOLUTION_BITS) & BIN_DEG_MASK)) : \
     (  ((d) >> BIN_DEG_SUB_RESOLUTION_BITS) & BIN_DEG_MASK))
#define BinDeg2HDegN(d) ((d) << BIN_DEG_SUB_RESOLUTION_BITS)


// 2進角度d0とdの和を返す
// ※2進方位は0以上BIN_DEG_RESOLUTION未満なので、結果が範囲外になる場合は折り返す
//   例: BinDegAdd(30, 50)...80, BinDegAdd(210, 60)...14
#define BinDegAdd(d0, d) \
    (((d0) + (d)) & BIN_DEG_MASK)
#define BinDegSub(d0, d)                              \
    BinDegAdd(d0, -d)
//    (((d0) - (d) + BIN_DEG_RESOLUTION) & BIN_DEG_MASK)
#define BinHDegAdd(d0, d) \
    (((d0) + (d)) & BIN_HDEG_MASK)
#define BinHDegSub(d0, d)                              \
    BinHDegAdd(d0, -d)

#define BinDegComplementAngle(d) (BinDegSub(BIN_DEG_RANGLE, (d)))
#define BinDegSupplementAngle(d) (BinDegSub((BIN_DEG_SANGLE), (d)))
#define BinDegNegativeAngle(d) (BinDegSub(BIN_DEG_RESOLUTION, (d)))

#define BinDegSinN(d) (_bin_deg_sin_table[d])
#define BinDegCosN(d) (_bin_deg_sin_table[BinDegAdd(d, BIN_DEG_RANGLE)])
//#define BinDegCosN(d) (_bin_deg_cos_table[d])
#define BinDegTanN(d) (_bin_deg_tan_table[d])
#define BinDegAtan(y, x) (bin_deg_atan2(y, x))

#define BinDegSin(d) BinDegSinN(BinDegNormalize(d))
#define BinDegCos(d) BinDegCosN(BinDegNormalize(d))
#define BinDegTan(d) BinDegTanN(BinDegNormalize(d))

int bin_deg_init(void);
int bin_deg_atan2(double y, double x);

#endif

